package factory.manage_orders;

import javax.ejb.*;
import javax.naming.*;
import javax.rmi.PortableRemoteObject;
import java.rmi.RemoteException;
import java.util.Date;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import factory.order.OrderHome;
import factory.order.Order;
import factory.order.OrderPK;
import factory.order.StatusStrings;
import factory.product.Product;
import factory.product.ProductHome;
import factory.product.RoutingInstruction;
import factory.order.OrderNotCancelableException;

public class ManageOrdersEJB implements SessionBean
{
  private static final int MILLIS_IN_DAYS = 86400000;

  // Business methods

  public void placeOrder(int salesDivision, int orderNumber,
                         String productName, Date dateDue)
              throws NoSuchProductException, DuplicateOrderException
  {
    try
    {
      // Find the product
      ProductHome productHome = getProductHome();
      Product product = productHome.findByPrimaryKey(productName);
      // create the order
      OrderHome orderHome = getOrderHome();
      orderHome.create(salesDivision, orderNumber, product, dateDue);
    }
    catch(NamingException ne)
    {
      throw new EJBException(ne);
    }
    catch(RemoteException re)
    {
      throw new EJBException(re);
    }
    catch (FinderException fe)
    {
      if(fe instanceof ObjectNotFoundException)
         throw new NoSuchProductException();
      else
         throw new EJBException(fe);
    }
    catch(CreateException ce)
    {
      if(orderExists(salesDivision, orderNumber))
         throw new DuplicateOrderException();
      else
         throw new EJBException(ce);
    }
  }

  public void cancelOrder(int salesDivision, int orderNumber)
              throws NoSuchOrderException, OrderNotCancelableException
  {
    try
    {
      // Find the order
      OrderHome orderHome = getOrderHome();
      OrderPK orderPK = new OrderPK(salesDivision, orderNumber);

      // Cancel it
      Order order = orderHome.findByPrimaryKey(orderPK);
      order.cancelOrder();
    }
    catch(NamingException ne)
    {
      throw new EJBException(ne);
    }
    catch(RemoteException re)
    {
      throw new EJBException(re);
    }
    catch(FinderException fe)
    {
      if(fe instanceof ObjectNotFoundException)
         throw new NoSuchOrderException();
      else
         throw new EJBException(fe);
    }
  }

  public OverdueOrderView[] getOverdueOrders()
  {
    try
    {
      LinkedList overdueOrders = new LinkedList();
      Date today = new Date();
      long todayMillis = today.getTime();
      long leadTimeMillis = getLeadTimeDays() * MILLIS_IN_DAYS;
      OrderHome orderHome = getOrderHome();
      Collection uncompletedOrders = orderHome.findUncompletedOrders();
      Iterator iterUncompletedOrders = uncompletedOrders.iterator();
      while(iterUncompletedOrders.hasNext())
      {
        Order uncompletedOrder = (Order)
              PortableRemoteObject.narrow(iterUncompletedOrders.next(),
                                          Order.class);
        Date dateDue = uncompletedOrder.getDateDue();
        String status = uncompletedOrder.getStatus();
        long dueDateMillis = dateDue.getTime();
        if(
           (status.equals(StatusStrings.OPEN) &&
              (todayMillis + leadTimeMillis > dueDateMillis))
                ||
            (status.equals(StatusStrings.IN_PROCESS) &&
              (todayMillis > dueDateMillis))
          )
        {
          OverdueOrderView view = new OverdueOrderView(
                           uncompletedOrder.getSalesDivision(),
                           uncompletedOrder.getOrderNumber(),
                           uncompletedOrder.getProductOrdered().getName(),
                           status, dateDue);
          overdueOrders.add(view);
        }
      }
      OverdueOrderView[] overdue = new OverdueOrderView[  
                                       overdueOrders.size()];
      return (OverdueOrderView[]) overdueOrders.toArray(overdue);
    }
    catch(NamingException ne)
    {
      throw new EJBException(ne);
    }
    catch(RemoteException re)
    {
      throw new EJBException(re);
    }
    catch(FinderException fe)
    {
      throw new EJBException(fe);
    }
  }

  public OpenOrderView[] getSchedulableOrders()
  {
    try
    {
      LinkedList schedulableOrders = new LinkedList();
      Date today = new Date();
      long todayMillis = today.getTime();
      long maxInventoryTimeMillis = getMaxInventoryTimeDays() * 
                                    MILLIS_IN_DAYS;
      long leadTimeMillis = getLeadTimeDays() * MILLIS_IN_DAYS;
      OrderHome orderHome = getOrderHome();
      Collection openOrders = orderHome.findOpenOrders();
      Iterator iterOpenOrders = openOrders.iterator();
      while(iterOpenOrders.hasNext())
      {
        Order openOrder = (Order)
          PortableRemoteObject.narrow(iterOpenOrders.next(),
                                      Order.class);
        Date dateDue = openOrder.getDateDue();
        long dueDateMillis = dateDue.getTime();
        if(todayMillis >= dueDateMillis - leadTimeMillis - maxInventoryTimeMillis)
         {
          OpenOrderView view = new OpenOrderView(
                        openOrder.getSalesDivision(),
                        openOrder.getOrderNumber(),
                        openOrder.getProductOrdered().getName(),
                        dateDue);
          schedulableOrders.add(view);
        }
      }
      OpenOrderView[] schedulable =
                      new OpenOrderView[schedulableOrders.size()];
      return (OpenOrderView[]) schedulableOrders.toArray(schedulable);
    }
    catch(NamingException ne)
    {
      throw new EJBException(ne);
    }
    catch(RemoteException re)
    {
      throw new EJBException(re);
    }
    catch(FinderException fe)
    {
      throw new EJBException(fe);
    }
  }

  public void createSampleProducts()
  {
    try
    {
      ProductHome productHome = getProductHome();
      // Create three sample products
      ProductCreationHelper pch = new ProductCreationHelper(productHome);
      pch.createAll();
    }
    catch(NamingException ne)
    {
      throw new EJBException(ne);
    }
  }

  public void createProduct(String id, String name)
  {
    try
    {
      ProductHome productHome = getProductHome();
      productHome.create(id, name, new RoutingInstruction[]{});
    }
    catch(NamingException ne)
    {
      throw new EJBException(ne);
    }
    catch(CreateException ce)
    {
      throw new EJBException(ce);
    }
    catch (RemoteException re)
    {
      throw new EJBException(re);
    }
  }

  public void addRoutingInstruction(String id, int sequence,
                                    String instruction)
  {
    try
    {
      ProductHome productHome = getProductHome();
      Product product = productHome.findByPrimaryKey(id);
      product.addRoutingInstruction(sequence, instruction);
    }
    catch(FinderException fe)
    {
      throw new EJBException(fe);
    }
    catch(NamingException ne)
    {
      throw new EJBException(ne);
    }
    catch(RemoteException re)
    {
      throw new EJBException(re);
    }
  }
// Implementation helpers

  private boolean orderExists(int salesDivision, int orderNumber)
  {
    try
    {
      OrderHome orderHome = getOrderHome();
      OrderPK orderPK = new OrderPK(salesDivision, orderNumber);
      Order order = orderHome.findByPrimaryKey(orderPK);
      return true;
    }
    catch(Exception e)
    {
      return false;
    }
  }

  private int getLeadTimeDays() throws NamingException
  {
    InitialContext initial = new InitialContext();
    Integer leadTimeDays = (Integer) 
                           initial.lookup("java:comp/env/lead_time");
    // A null pointer will roll back the transaction
    return leadTimeDays.intValue();
  }

  private int getMaxInventoryTimeDays() throws NamingException
  {
    InitialContext initial = new InitialContext();
    Integer inventoryTimeDays = (Integer)initial.lookup(
                                "java:comp/env/max_inventory_time");
    // A null pointer will roll back the transaction
    return inventoryTimeDays.intValue();
  }

  private ProductHome getProductHome() throws NamingException
  {
    InitialContext initial = new InitialContext();
    ProductHome home = (ProductHome)
                javax.rmi.PortableRemoteObject.narrow(
                          initial.lookup("java:comp/env/ejb/Product"), 
                                         ProductHome.class );
    return home;
  }

  private OrderHome getOrderHome() throws NamingException
  {
    InitialContext initial = new InitialContext();
    OrderHome home = (OrderHome)javax.rmi.PortableRemoteObject.narrow(
                                initial.lookup("java:comp/env/ejb/Order"), 
                                               OrderHome.class );
    return home;
  }
// Framework & lifecycle methods

  public void ejbCreate()
  {
  }

  public void ejbActivate()
  {
  }

  public void ejbPassivate()
  {
  }

  public void ejbRemove()
  {
  }

  public void setSessionContext(SessionContext ctx)
  {
  }
}
